"""Provides the `Transformer` for correcting bleedthrough images."""

import json
import pathlib

import numpy
import scipy.ndimage

__all__ = ["Transformer"]


class Transformer:
    """A thread-save transformer created from a trained Theia model."""

    def __init__(
        self,
        *,
        num_channels: int,
        channel_overlap: int,
        beta: float,
        contribution_kernels: dict[tuple[int, int], numpy.ndarray],
        interactions_kernels: dict[tuple[int, int], numpy.ndarray],
    ) -> None:
        """This is meant to only be created by Theia.

        Args:
            num_channels: Same as with Theia.
            channel_overlap: Same as with Theia.
            beta: Same as with Theia.
            contribution_kernels: Learned in Theia.
            interactions_kernels: Learned in Theia.
        """
        self._num_channels = num_channels
        self._channel_overlap = channel_overlap
        self._beta = beta
        self._contribution_kernels = contribution_kernels
        self._interactions_kernels = interactions_kernels

    def save(self, json_path: pathlib.Path) -> None:
        """Save the Transformer as a JSON file."""
        params = {
            "num_channels": self._num_channels,
            "channel_overlap": self._channel_overlap,
            "beta": self._beta,
            "contribution_kernels": {
                f"{i}-{j}": self._kernel_to_list(v)
                for (i, j), v in self._contribution_kernels.items()
            },
            "interactions_kernels": {
                f"{i}-{j}": self._kernel_to_list(v)
                for (i, j), v in self._interactions_kernels.items()
            },
        }
        with json_path.open("w") as writer:
            json.dump(params, writer, indent=2)

    @staticmethod
    def _kernel_to_list(kernel: numpy.ndarray) -> list[float]:
        return list(map(float, kernel.flatten()))

    @staticmethod
    def load(json_path: pathlib.Path) -> "Transformer":
        """Load the Transformer from a saved json."""
        with json_path.open("r") as reader:
            params = json.load(reader)
        params["contribution_kernels"] = {
            tuple(map(int, k.split("-"))): Transformer._list_to_kernel(v)
            for k, v in params["contribution_kernels"].items()
        }
        params["interactions_kernels"] = {
            tuple(map(int, k.split("-"))): Transformer._list_to_kernel(v)
            for k, v in params["interactions_kernels"].items()
        }
        return Transformer(**params)

    @staticmethod
    def _list_to_kernel(kernel: list[float]) -> numpy.ndarray:
        h = int(numpy.sqrt(len(kernel)))
        kernel_array = numpy.asarray(kernel, numpy.float32)
        return numpy.reshape(kernel_array, newshape=(h, h))

    @property
    def num_channels(self) -> int:
        """The number of channels in each image."""
        return self._num_channels

    @property
    def channel_overlap(self) -> int:
        """The number of adjacent channels to use for bleed-through estimation."""
        return self._channel_overlap

    @property
    def contribution_kernels(self) -> dict[tuple[int, int], numpy.ndarray]:
        """Return the fitted contribution kernels."""
        if len(self._contribution_kernels) == 0:
            message = "Please call `fit` before using this property."
            raise ValueError(message)
        return self._contribution_kernels

    @property
    def interaction_kernels(self) -> dict[tuple[int, int], numpy.ndarray]:
        """Return the fitted interaction kernels."""
        if len(self._interactions_kernels) == 0:
            message = "Please call `fit` before using this property."
            raise ValueError(message)
        return self._interactions_kernels

    def bleedthrough_components(self, image: numpy.ndarray) -> numpy.ndarray:
        """Compute the channel-wise bleedthrough components for the image."""
        bleedthrough = numpy.stack(
            [numpy.zeros_like(image) for _ in range(self.num_channels)],
            axis=-1,
        )

        for i_target in range(self.num_channels):
            neighbor_indices = self._neighbor_indices(i_target)

            neighbors = [image[:, :, i] for i in neighbor_indices]
            kernels = [
                self.contribution_kernels[(i_target, i)] for i in neighbor_indices
            ]

            bleedthrough[:, :, i_target, :] = self._compute_contributions(
                neighbor_indices,
                neighbors,
                kernels,
            )

        return bleedthrough

    def total_bleedthrough(self, image: numpy.ndarray) -> numpy.ndarray:
        """Compute the total bleedthrough components for the image."""
        bleedthrough = self.bleedthrough_components(image)
        return numpy.sum(bleedthrough, axis=-1)

    def interactions_components(self, image: numpy.ndarray) -> numpy.ndarray:
        """Compute the interaction components for the image."""
        bleedthrough = numpy.stack(
            [numpy.zeros_like(image) for _ in range(self.num_channels)],
            axis=-1,
        )

        for i_target in range(self.num_channels):
            target = image[:, :, i_target]

            neighbor_indices = self._neighbor_indices(i_target)
            interactions = [
                self._compute_interaction(target, image[:, :, i])
                for i in neighbor_indices
            ]
            kernels = [
                self.interaction_kernels[(i_target, i)] for i in neighbor_indices
            ]

            bleedthrough[:, :, i_target, :] = self._compute_contributions(
                neighbor_indices,
                interactions,
                kernels,
            )

        return bleedthrough

    def total_interactions(self, image: numpy.ndarray) -> numpy.ndarray:
        """Compute the total interaction components for the image."""
        interactions = self.interactions_components(image)
        return numpy.sum(interactions, axis=-1)

    def transform(
        self,
        image: numpy.ndarray,
        *,
        remove_interactions: bool = False,
    ) -> numpy.ndarray:
        """Transform and return a multichannel image.

        Args:
            image: A multichannel image with shape (H, W, C) where H is the
             height, W is the width and C is the number of channels.
            remove_interactions: Whether to remove interaction components.

        Returns:
            The corrected image.
        """
        bleed_through = self.total_bleedthrough(image)

        if remove_interactions:
            bleed_through += self.total_interactions(image)

        return numpy.clip(image - bleed_through, a_min=0.0)

    def _neighbor_indices(self, i_target: int) -> list[int]:
        min_i = max(i_target - self.channel_overlap, 0)
        max_i = min(i_target + self.channel_overlap, self.num_channels)
        return [i for i in range(min_i, max_i) if i != i_target]

    def _compute_interaction(
        self,
        target: numpy.ndarray,
        neighbor: numpy.ndarray,
    ) -> numpy.ndarray:
        interaction = numpy.power(neighbor, self._beta)
        interaction = numpy.multiply(target, interaction)
        interaction = numpy.power(interaction, 1 / (1 + self._beta))
        return interaction

    def _compute_contributions(
        self,
        indices: list[int],
        neighbors: list[numpy.ndarray],
        kernels: list[numpy.ndarray],
    ) -> numpy.ndarray:
        correlations = numpy.zeros(shape=(*neighbors[0].shape, self.num_channels))
        for i, n, k in zip(indices, neighbors, kernels):
            # noinspection PyUnresolvedReferences
            correlations[:, :, i] = scipy.ndimage.correlate(n, k)
        return correlations



"""Theia implemented using a CNN in Tensorflow."""
import pathlib
import typing

import numpy
import tensorflow
from tensorflow.python import keras
from tensorflow.python.keras import callbacks
from tensorflow.python.keras import layers
from tensorflow.python.keras import metrics

from ..data import TileGenerator
from .base import Theia

__all__ = ["Neural"]


class Neural(Theia, keras.Model):  # type: ignore
    """Theia implemented our a Neural Network Architecture.

    The architecture is a generalization of Siamese Networks, and performs the
     equivalent of LASSO regression.
    """

    def __init__(
        self,
        *,
        num_channels: int,
        channel_overlap: int,
        kernel_size: int,
        alpha: float,
        beta: float,
        tile_size: int,
    ) -> None:
        """Initialize and build the Network.

        Args:
            num_channels: Number of channels in each input image.
            channel_overlap: Maximum number of adjacent channels to consider for
             bleed-through removal.
            kernel_size: Side-length of square kernel (convolutional) to use for
             estimating bleed-through.
            alpha: Relative size of l1-penalty in the LASSO loss.
            beta: Relative weighting of target channel in interaction terms.
            tile_size: Side-length of square tiles to use as inputs to the
             network.
        """
        keras.Model.__init__(self)
        Theia.__init__(
            self,
            num_channels=num_channels,
            channel_overlap=channel_overlap,
            kernel_size=kernel_size,
            alpha=alpha,
            beta=beta,
        )

        self._callbacks: list[callbacks.Callback] = []

        self._tile_shape = (tile_size, tile_size, 1)
        self._sources = [
            layers.Input(self._tile_shape, name=f"source_{i}")
            for i in range(num_channels)
        ]

        self._kernel_names = []
        self._num_kernels: list[int] = []
        self._contributions: list[list[_TheiaConv]] = []
        self._aggregators: list[_Aggregator] = []

        for source_index, source in enumerate(self._sources):
            contributions: list[_TheiaConv] = []

            for offset in range(-self._channel_overlap, self._channel_overlap + 1):
                if offset == 0:
                    continue

                neighbor_index = source_index + offset
                if neighbor_index not in range(self._num_channels):
                    continue

                neighbor = self._sources[neighbor_index]
                contribution = _TheiaConv(
                    name=f"contribution_{source_index}_{neighbor_index}",
                    kernel_size=self._kernel_size,
                )(neighbor)
                self._kernel_names.append((True, source_index, neighbor_index))
                contributions.append(contribution)

                interaction_term = tensorflow.pow(neighbor, self._beta)
                interaction_term = tensorflow.multiply(interaction_term, source)
                interaction_term = tensorflow.pow(
                    interaction_term,
                    1 / (1 + self._beta),
                )
                interaction = _TheiaConv(
                    name=f"interactions_{source_index}_{neighbor_index}",
                    kernel_size=self._kernel_size,
                )(interaction_term)
                self._kernel_names.append((False, source_index, neighbor_index))
                contributions.append(interaction)

            self._num_kernels.append(len(contributions))
            self._contributions.append(contributions)
            self._aggregators.append(
                _Aggregator(name=f"aggregation_{source_index}")(
                    [source, *contributions],
                ),
            )

        self.inputs = self._sources
        self.outputs = self._aggregators

        self._model = keras.Model(
            inputs=self.inputs,
            outputs=self.outputs,
        )
        self.loss_tracker = metrics.Mean(name="theia_loss")

        super().build(
            [
                (None, self.tile_size, self.tile_size, 1)
                for _ in range(self.num_channels)
            ],
        )

    @property
    def tile_size(self) -> int:
        """Side-length of square tiles to use as inputs to the network."""
        return self._tile_shape[0]

    def fit_theia(  # type: ignore
        self,
        *,
        train_gen: TileGenerator,
        valid_gen: TileGenerator,
        epochs: int,
        verbose: int,
    ) -> None:
        """Fit Theia to a multichannel image.

        Args:
            train_gen: The tiles to train on.
            valid_gen: The tiles to validate on.
            epochs: The number of epochs to train the model.
            verbose: verbosity mode.
        """
        self.fit(
            x=train_gen,
            epochs=epochs,
            verbose=verbose,
            callbacks=self._callbacks,
            validation_data=valid_gen,
        )

        random_inputs = [
            numpy.random.random((1, *self._tile_shape))
            for _ in range(self._num_channels)
        ]
        outputs = self(random_inputs)

        kernels: numpy.ndarray = numpy.squeeze(
            numpy.stack(
                [k for out in outputs for k in tensorflow.unstack(out[-1])],
                axis=0,
            ),
        ).astype(numpy.float32)
        named_kernels = {
            name: kernels[i, ...] for i, name in enumerate(self._kernel_names)
        }

        self._contribution_kernels = {
            (i, j): k for (b, i, j), k in named_kernels.items() if b
        }

        self._interactions_kernels = {
            (i, j): k for (b, i, j), k in named_kernels.items() if not b
        }

    def early_stopping(
        self,
        *,
        min_delta: float,
        patience: int,
        verbose: int,
        restore_best_weights: bool,
    ) -> None:
        """Set up an EarlyStopping callback for the model.

        Args:
            min_delta: Minimum change in the monitored quantity to qualify as an
             improvement, i.e. an absolute change of less than min_delta, will
             count as no improvement.
            patience: Number of epochs with no improvement after which training
             will be stopped.
            verbose: verbosity mode.
            restore_best_weights: Whether to restore model weights from the
             epoch with the best value of the loss function.
        """
        self._callbacks.append(
            callbacks.EarlyStopping(
                monitor="val_theia_loss",
                min_delta=min_delta,
                patience=patience,
                verbose=verbose,
                mode="min",
                restore_best_weights=restore_best_weights,
            ),
        )

    def add_callback(self, cb: callbacks.Callback) -> None:  # noqa
        self._callbacks.append(cb)

    def save(self, path: pathlib.Path) -> None:
        """Save the model to the given `path`."""
        raise NotImplementedError

    @staticmethod
    def load(path: pathlib.Path) -> "Neural":
        """Load the model from the given `path`."""
        raise NotImplementedError

    def call(
        self,
        inputs: list[tensorflow.Tensor],
        training: typing.Optional[typing.Union[bool, tensorflow.Tensor]] = None,
        mask: typing.Optional[
            typing.Union[tensorflow.Tensor, list[tensorflow.Tensor]]
        ] = None,
    ) -> list[tuple[tensorflow.Tensor, tensorflow.Tensor]]:
        """Call the model."""
        return self._model(inputs)  # type: ignore

    def train_step(self, inputs: list[tensorflow.Tensor]) -> dict[str, float]:
        """Take one step of training the model."""
        with tensorflow.GradientTape() as tape:
            losses = self._compute_loss(inputs)

        trainable_weights = self._model.trainable_weights
        gradients = tape.gradient(losses, trainable_weights)
        self.optimizer.apply_gradients(zip(gradients, trainable_weights))

        self.loss_tracker.update_state(tensorflow.reduce_mean(losses))
        return {self.loss_tracker.name: self.loss_tracker.result()}

    def test_step(self, inputs: list[tensorflow.Tensor]) -> dict[str, float]:
        """Take one step of testing/validating the model."""
        losses = self._compute_loss(inputs)
        self.loss_tracker.update_state(tensorflow.reduce_mean(losses))
        return {self.loss_tracker.name: self.loss_tracker.result()}

    def _compute_loss(self, inputs: list[tensorflow.Tensor]) -> list[tensorflow.Tensor]:
        losses = []

        for n, (corrected, kernels) in zip(self._num_kernels, self._model(inputs)):
            sq_error = tensorflow.reduce_mean(tensorflow.square(corrected))

            kernel_stack = tensorflow.abs(tensorflow.stack(kernels))
            l1_penalty = self._alpha * tensorflow.reduce_mean(kernel_stack)

            loss = tensorflow.maximum(sq_error + l1_penalty, 0.0)
            losses.extend([loss] * n)

        return losses

    @property
    def metrics(self) -> list[metrics.Metric]:
        """Metric to track."""
        return [self.loss_tracker]

    def get_config(self) -> None:  # noqa
        raise NotImplementedError

    def _serialize_to_tensors(self) -> None:
        raise NotImplementedError

    def _restore_from_tensors(
        self,
        restored_tensors: dict[str, tensorflow.Tensor],
    ) -> None:
        raise NotImplementedError


class _TheiaConv(layers.Conv2D):  # type: ignore
    def __init__(
        self,
        *,
        name: str,
        kernel_size: int,
    ) -> None:
        super().__init__(
            name=name,
            filters=1,
            padding="same",
            kernel_size=kernel_size,
            use_bias=False,
            kernel_constraint=tensorflow.keras.constraints.NonNeg(),
        )

    def call(
        self,
        bleedthrough_term: tensorflow.Tensor,
    ) -> tuple[tensorflow.Tensor, tensorflow.Tensor]:
        conv = super().call(bleedthrough_term)
        return conv, self.kernel

    def _serialize_to_tensors(self) -> None:
        raise NotImplementedError

    def _restore_from_tensors(
        self,
        restored_tensors: dict[str, tensorflow.Tensor],
    ) -> None:
        raise NotImplementedError


class _Aggregator(layers.Layer):  # type: ignore
    def __init__(self, *, name: str) -> None:
        super().__init__(name=name)

    def get_config(self) -> typing.Any:  # noqa
        return super().get_config()

    # noinspection PyMethodOverriding
    def call(
        self,
        inputs: list[tensorflow.Tensor],
    ) -> tuple[tensorflow.Tensor, tensorflow.Tensor]:
        [source, *contributors] = inputs

        kernels = tensorflow.stack([k for _, k in contributors])
        filtered = [f for f, _ in contributors]

        bleed_through = tensorflow.add_n(filtered)

        corrected = tensorflow.subtract(source, bleed_through)
        return corrected, kernels

    def _serialize_to_tensors(self) -> None:
        raise NotImplementedError

    def _restore_from_tensors(
        self,
        restored_tensors: dict[str, tensorflow.Tensor],
    ) -> None:
        raise NotImplementedError



"""A Tiled Image Data Generator for Theia."""
import random

import numpy
import tensorflow
from tensorflow.python.keras.utils import data_utils as keras_utils

from .utils import constants

__all__ = ["TileGenerator"]


class TileGenerator(keras_utils.Sequence):  # type: ignore
    """Take images as numpy arrays and present tiles to the Neural Network."""

    def __init__(
        self,
        *,
        images: list[numpy.ndarray],
        tile_size: int,
        shuffle: bool = True,
        normalize: bool = True,
    ) -> None:
        """Create a Tile Generator from the given images.

        Let `n` be the number of images, `w` by the width, `h` be the height and
         `c` be the number of channels. `images` must have `n` items and each
         should be an image with shape `(h, w, c)`.

        Args:
            images: to correct for bleed-through.
            tile_size: If `w` or `h` is larger than `tile_size` then the images
             will be tiled. All tiles will be padded to a square of side-length
             `tile_size`.
            shuffle: Whether to shuffle tiles after each epoch.
            normalize: Whether to normalize the image intensities to a [0, 1]
             range.
        """
        [h, w, c] = list(map(int, images[0].shape))
        h_pad, self._h_tiles = _calculate_padding(h, tile_size)
        w_pad, self._w_tiles = _calculate_padding(w, tile_size)
        self._paddings = tensorflow.constant([[0, h_pad], [0, w_pad], [0, 0]])

        if normalize:
            images = list(map(_normalize_image, images))
        else:
            images = [img / (numpy.max(img) + constants.EPSILON) for img in images]
        self._images = [
            tensorflow.convert_to_tensor(image, dtype=tensorflow.float32)
            for image in images
        ]

        self._tile_shape = (tile_size, tile_size, c)
        self._shuffle = shuffle

        self.on_epoch_end()

    @property
    def num_channels(self) -> int:
        """The number of channels in each image."""
        return self._tile_shape[-1]

    @property
    def tile_shape(self) -> tuple[int, int, int]:
        """The shape of tiles: `(h, w, c)`."""
        return self._tile_shape

    def on_epoch_end(self) -> None:
        """Optionally shuffle images after each epoch."""
        if self._shuffle:
            random.shuffle(self._images)

    def __getitem__(self, index: int) -> list[tensorflow.Tensor]:
        """Return a batch of tiles.

        Args:
            index: of the image from which to make tiles.

        Returns:
            A list with `c` `Tensors`, each with shape
             `(num_tiles, tile_size, tile_size, 1)`
        """
        image = tensorflow.pad(self._images[index], self._paddings)
        tiled_image: tensorflow.Tensor = tensorflow.reshape(
            image,
            (self._h_tiles, self._w_tiles, *self._tile_shape),
        )
        tiled_image = tensorflow.reshape(
            tiled_image,
            (-1, *self._tile_shape),
        )
        batch: list[tensorflow.Tensor] = tensorflow.unstack(tiled_image, axis=-1)
        return batch

    def __len__(self) -> int:
        """Number of images."""
        return len(self._images)


def _normalize_image(image: numpy.ndarray) -> numpy.ndarray:
    min_val = numpy.min(image, axis=(0, 1))
    max_val = numpy.max(image, axis=(0, 1))
    image = image.astype(numpy.float32) - min_val
    image /= max_val - min_val + constants.EPSILON
    return image


def _calculate_padding(stop: int, step: int) -> tuple[int, int]:
    if stop <= step:
        return stop - step, 1

    padding = stop % step
    n_tiles = stop // step
    if padding == 0:
        return padding, n_tiles

    return padding, n_tiles + 1



"""The Abstract Base Class for Theia models."""

import abc
import pathlib
import typing

import numpy

from .transformer import Transformer


class Theia(abc.ABC):
    """The base model for Theia.

    All subclasses must implement `save`, `load`, `fit` and `transform`.
    """

    def __init__(
        self,
        *,
        num_channels: int,
        channel_overlap: int,
        kernel_size: int,
        alpha: float,
        beta: float,
    ) -> None:
        """Base class for a Theia model.

        Args:
            num_channels: number of channels in each image.
            channel_overlap: Maximum number of adjacent channels to consider for
             bleed-through removal.
            kernel_size: Side-length of square kernel (convolutional) to use for
             estimating bleed-through.
            alpha: Relative size of l1-penalty in the LASSO loss.
            beta: Relative weighting of target channel in interaction terms.
        """
        self._num_channels = num_channels
        self._channel_overlap = channel_overlap
        self._kernel_size = kernel_size
        self._alpha = alpha
        self._beta = beta

        self._contribution_kernels: dict[tuple[int, int], numpy.ndarray] = {}
        self._interactions_kernels: dict[tuple[int, int], numpy.ndarray] = {}

    @property
    def num_channels(self) -> int:
        """The number of channels in each image."""
        return self._num_channels

    @property
    def channel_overlap(self) -> int:
        """The number of adjacent channels to use for bleed-through estimation."""
        return self._channel_overlap

    @abc.abstractmethod
    def fit_theia(
        self,
        *args: list[typing.Any],
        **kwargs: dict[str, typing.Any],
    ) -> None:
        """Fit Theia to a multi-channel image."""
        pass

    @abc.abstractmethod
    def save(self, path: pathlib.Path) -> None:
        """Save the model to the given `path`."""
        pass

    @staticmethod
    @abc.abstractmethod
    def load(path: pathlib.Path) -> "Theia":
        """Load the model from the given `path`."""
        pass

    @property
    def contribution_kernels(self) -> dict[tuple[int, int], numpy.ndarray]:
        """Return the fitted contribution kernels."""
        if len(self._contribution_kernels) == 0:
            message = "Please call `fit` before using this property."
            raise ValueError(message)
        return self._contribution_kernels

    @property
    def interactions_kernels(self) -> dict[tuple[int, int], numpy.ndarray]:
        """Return the fitted interaction kernels."""
        if len(self._interactions_kernels) == 0:
            message = "Please call `fit` before using this property."
            raise ValueError(message)
        return self._interactions_kernels

    @property
    def transformer(self) -> "Transformer":
        """Provide a thread-safe transformer to apply bleed-through correction."""
        return Transformer(
            num_channels=self.num_channels,
            channel_overlap=self.channel_overlap,
            beta=self._beta,
            contribution_kernels=self.contribution_kernels,
            interactions_kernels=self.interactions_kernels,
        )



"""Expose basics of Theia."""

from .data import TileGenerator
from .models import Neural
from .models import Transformer
from .utils import constants

__all__ = ["Transformer", "Neural", "TileGenerator", "constants"]



"""Some helpful constants."""

import logging
import multiprocessing
import os

THEIA_LOG = getattr(logging, os.environ.get("THEIA_LOG_LEVEL", "INFO"))

MAX_WORKERS = max(1, multiprocessing.cpu_count() // 2)
MIN_DATA_SIZE = 2**19  # Load at least this much data
MAX_DATA_SIZE = 2**31  # Limit to loading 500MB of pixels
EPSILON = 1e-8  # To avoid divide-by-zero errors
